from sli_lib.fodot.vocabulary import Vocabulary, BuiltinTypes
from sli_lib.fodot.structure import StrInterp
import pytest

def test_add_type():
    vocab = Vocabulary("V")
    vocab.add_type("T1")
    vocab.add_type("T2")
    T1 = vocab.parse_type("T1")
    vocab.add_voc_type_interp(T1, StrInterp(["a", "b", "c"]))
    with pytest.raises(ValueError):
        vocab.add_voc_type_interp("T2", StrInterp(["a", "b", "c"]))
    vocab2 = Vocabulary("V")
    vocab2.add_type("T2")
    T12 = vocab2.parse_type("T2")
    with pytest.raises(ValueError):
        vocab.add_voc_type_interp(T12, StrInterp(["d", "e", "f"]))
    vocab.add_voc_type_interp("T2", StrInterp(["d", "e", "f"]))

def test_add_pfunc():
    vocab = Vocabulary("V")
    vocab.add_type("T")
    vocab.add_pfunc("p", (), "Bool")
    with pytest.raises(TypeError):
        vocab.add_pfunc(2, (), "Bool") # type: ignore
    with pytest.raises(TypeError):
        vocab.add_pfunc(("ttt", 4), (), "Bool")  # type: ignore
    vocab.add_pfunc((f"p{i}" for i in range(5)), (), "Int")
    vocab.add_pfunc((f"t{i}" for i in range(5)), ("T" for _ in range(3)), "Int")
    T = vocab.parse_type("T")
    vocab.add_pfunc("w", (T, T), BuiltinTypes.INT)
    vocab2 = Vocabulary("T")
    vocab2.add_type("T")
    T2 = vocab2.parse_type("T")
    with pytest.raises(ValueError):
        vocab.add_pfunc("T", (T2,), "Bool")
    vocab.add_pfunc("o", (), BuiltinTypes.BOOL)



from ..structure import Structure
from ..vocabulary import Vocabulary
from typing import Self

class Assertions:
    def __init__(self, vocabulary: Vocabulary) -> None: ...
    def vocab(self) -> Vocabulary: ...
    def parse(self, decls: str) -> Self: ...
    def __str__(self) -> str: ...

class Theory:
    def __init__(self, assertions: Assertions, structure: Structure) -> None: ...
    @classmethod
    def from_specification(cls, spec: str) -> Self: ...
    def structure(self) -> Structure: ...
    def vocab(self) -> Vocabulary: ...
    def assertions(self) -> Assertions: ...
    def __str__(self) -> str: ...



from ..._fodot.vocabulary import *
from ..._fodot import vocabulary

# Don't enumerate each thing we want to import here do it in __init__.pyi
from ._pure import *
from . import _pure
from ..._all_utils import filter__all__, merge__all__, rename_module

__all__ = merge__all__(
    filter__all__(vocabulary.__all__),
    _pure.__all__,
)

# patch __module__ of everything to this module
rename_module(__all__, locals(), __name__)



from __future__ import annotations
from ._pure import *
from ..structure import TypeInterp
from typing import Iterable, Self

class Symbol:
    def name(self) -> str: ...
    def __str__(self) -> str: ...
class CustomSymbol(Symbol):
    def vocab(self) -> "Vocabulary": ...
class BuiltinBool(Symbol): ...
class BuiltinInt(Symbol): ...
class BuiltinReal(Symbol): ...
class IntType(CustomSymbol): ...
class RealType(CustomSymbol): ...
class StrType(CustomSymbol): ...
class Pfunc(CustomSymbol): ...

class TypeIter:
    def __iter__(self) -> Self: ...
    def __next__(self) -> CustomType: ...
    def vocab(self) -> "Vocabulary": ...

class PfuncIter:
    def __iter__(self) -> Self: ...
    def __next__(self) -> Pfunc: ...
    def vocab(self) -> "Vocabulary": ...

class Real:
    def __init__(self, value: int | float | str): ...
    def __add__(self, other: Self) -> Self: ...
    def __sub__(self, other: Self) -> Self: ...
    def __mul__(self, other: Self) -> Self: ...
    def __div__(self, other: Self) -> Self: ...
    def __mod__(self, other: Self) -> Self: ...
    def __float__(self) -> float: ...


class Vocabulary:
    def __init__(self, name: str): ...
    def name(self) -> str: ...
    def parse(self, decls: str) -> Self: ...
    def add_type(self, name: str, super_set: ExtType | None = None) -> None: ...
    def add_voc_type_interp(self, type: ExtCustomType, interp: TypeInterp) -> None: ...
    def parse_type(self, value: str) -> Type: ...
    def parse_pfunc(self, value: str) -> Pfunc: ...
    def parse_symbol(self, value: str) -> Symbol: ...
    def add_pfunc(
        self, name: str | Iterable[str], domain: Iterable[ExtType], codomain: ExtType
    ) -> None: ...
    def iter_pfuncs(self) -> PfuncIter: ...
    def iter_types(self) -> TypeIter: ...
    def __str__(self) -> str: ...



from sli_lib.fodot import Vocabulary, Structure
from mypy import api
from sys import stdout, stderr
import pytest

def run_in_mypy(to_run: str) -> tuple[str, str, int]:
    ret = api.run(["-c", to_run])
    return ret

def test_incomplete_ops():
    vocab = Vocabulary("V").parse(
        """
        type A := {1,2,3}
        p: A * A -> Bool
        t: A * A -> A
        """)
    structure = Structure(vocab)
    t_decl = vocab.parse_pfunc("t")
    t = structure["t"]
    t = structure[t_decl]
    try:
        t.set((1,1), 4)
    except ValueError:
        pass
    try:
        t.set((1,1), True)
    except ValueError:
        pass
    assert not t.any_known()
    t.set((1,1), 1)
    A = vocab.parse_type("A")
    assert t.get(1,1) == t(1,1) == 1
    assert t.amount_known() == 1
    assert t.amount_unknown() == len(structure.get_type_interp("A"))**2 - 1
    assert t.amount_unknown() == len(structure.get_type_interp(A))**2 - 1
    t_iter = iter(t)
    t_iter2 = iter(t)
    assert next(t_iter) == ((1,1), 1)
    assert next(t_iter2) == ((1,1), 1)
    with pytest.raises(StopIteration):
        next(t_iter)
    with pytest.raises(StopIteration):
        next(t_iter2)

def test_type_narrowing():
    to_run = """
from sli_lib.fodot.structure import (
    has_codomain, PfuncInterp, GlobModelPfuncInterp, ModelPfuncInterp,
    GlobModel, Model
)
from sli_lib.fodot import Structure
from typing import assert_type

def a(structure: Structure) -> None:
    p = structure["p"]
    if has_codomain(p, bool):
        reveal_type(p)
        assert_type(p, PfuncInterp[bool])

def b(glob_model: GlobModel) -> None:
    p = glob_model["p"]
    if has_codomain(p, bool):
        reveal_type(p)
        assert_type(p, GlobModelPfuncInterp[bool])

def c(model: Model) -> None:
    p = model["p"]
    if has_codomain(p, bool):
        reveal_type(p)
        assert_type(p, ModelPfuncInterp[bool])
"""
    out, err, exit_code = run_in_mypy(to_run)
    print(out, file=stdout)
    print(err, file=stderr)
    assert exit_code == 0



from ..fodot.theory import Theory
from ..fodot.structure import GlobModel, Structure, Model
from typing import Self
from ._pure import *

class Z3GlobModelIterator:
    def __iter__(self) -> Self: ...
    def __next__(self) -> GlobModel: ...

class Z3ModelIterator:
    def __iter__(self) -> Self: ...
    def __next__(self) -> Model: ...
    def enable_skip_infinite(self) -> Self: ...
    def disable_skip_infinite(self) -> Self: ...
    def skip_infinite(self, skip: bool) -> Self: ...

class Z3Solver:
    def __init__(self, theory: Theory): ...
    def theory(self) -> Theory: ...
    def check(self) -> SatResult: ...
    def get_model(self) -> GlobModel | None: ...
    def propagate(self) -> Structure | None: ...
    def iter_glob_models(self) -> Z3GlobModelIterator: ...
    def iter_models(self) -> Z3ModelIterator: ...



from ..._fodot.theory import *
from ..._fodot import theory
from ..._all_utils import filter__all__, rename_module

__all__ = filter__all__(theory.__all__)

# patch __module__ of everything to this module
rename_module(__all__, locals(), __name__)



from sli_lib.fodot import Theory
from sli_lib.solver import Z3Solver
import pytest

def test_z3_solver():
    theory  = Theory.from_specification("""
vocabulary {
    type T := { 0, 1 }
    p: T -> Bool
    t: T -> T
}
theory {
    !x in T: p(x) => t(x) = x.
}
structure {}
""")
    solver = Z3Solver(theory)
    assert solver.check()
    solver_iter = solver.iter_models()
    solver_iter.disable_skip_infinite()
    solver_iter.enable_skip_infinite()
    next(solver_iter)
    solver.check()
    with pytest.raises(RuntimeError):
        next(solver_iter)
    solver_iter = solver.iter_glob_models()
    for _ in solver_iter:
        pass
    assert not solver.check()



import pytest
from pdoc.doc import Module, Doc, Namespace
from typing import Generator, Iterable, NamedTuple, Self
import importlib
import sys


def iter_module(module: Module) -> Generator[Doc, None, None]:
    yield module
    for value in module.submodules:
        yield value
        yield from iter_module(value)
    for value in module.classes:
        yield value
        yield from iter_namespace(value)
    for value in module.functions:
        yield value
    for value in module.variables:
        yield value


def iter_namespace(namespace: Namespace) -> Generator[Doc, None, None]:
    for value in namespace.members.values():
        if isinstance(value, Namespace):
            yield from iter_namespace(value)
        else:
            yield value


class Doctest(NamedTuple):
    item: Doc
    doctest: str
    doctest_index: int

    @classmethod
    def gather_doctests(cls, doc: Doc) -> Generator[Self, None, None]:
        cur = 0
        count = 0
        BEGIN = "\n```python\n"
        END = "\n```"
        while True:
            begin = doc.docstring.find(BEGIN, cur)
            if begin < 0:
                break
            if begin + len(BEGIN) > len(doc.docstring):
                break
            end = doc.docstring.find(END, begin + len(BEGIN))
            if end < 0:
                break
            doctest = doc.docstring[begin + len(BEGIN) : end]
            cur = end
            yield cls(doc, doctest, count)
            count += 1

    def __str__(self) -> str:
        return f"{self.item.fullname}[{self.doctest_index}]"

    def __repr__(self) -> str:
        return f"<{self.item.fullname}[{self.doctest_index}]>"


def collect_docs(module: Module) -> Iterable[Doctest]:
    for item in iter_module(module):
        yield from Doctest.gather_doctests(item)


@pytest.mark.parametrize(
    "sli_lib_doc",
    collect_docs(Module(importlib.import_module("sli_lib"))),
    ids=map(
        lambda x: str(x),
        collect_docs(Module(importlib.import_module("sli_lib"))),
    ),
    scope="session",
)
def test_sli_lib_docs(sli_lib_doc: Doctest):
    print(sli_lib_doc.doctest, file=sys.stderr)
    exec(sli_lib_doc.doctest)



from .._solver import *
from .. import _solver
from ._pure import *
from .._all_utils import filter__all__, merge__all__, rename_module

__all__ = merge__all__(
    filter__all__(_solver.__all__),
    _pure.__all__
)

# patch __module__ of everything to this module
rename_module(__all__, locals(), __name__)



"""
Module containing methods and data structures for representing FO(·).

# Re-exports

- `.vocabulary.Vocabulary`
- `.structure.Structure`
- `.theory.Theory`
- `.theory.Assertions`

"""

from . import theory as theory
from . import structure as structure
from . import vocabulary as vocabulary

from .vocabulary import Vocabulary as Vocabulary
from .structure import Structure as Structure
from .theory import Theory as Theory, Assertions as Assertions

__all__ = [
    "theory",
    "structure",
    "vocabulary",
    "Vocabulary",
    "Structure",
    "Theory",
    "Assertions",
]



from sli_lib.fodot import Vocabulary, Assertions, Theory, Structure, structure
from sli_lib.solver import Z3Solver

def test_simple_theory():
    vocab = Vocabulary("V")
    vocab.add_type("A")
    vocab.add_pfunc("p", "A", "Bool")
    struct = Structure(vocab)
    struct.set_type_interp("A", structure.StrInterp(["a","b","c"]))
    assertions = Assertions(vocab)
    assertions.parse(
    """
        !x in A: p(x).
    """)
    theory = Theory(assertions, struct)
    z3_solver = Z3Solver(theory)
    assert z3_solver.check()
    model = z3_solver.get_model()
    assert model is not None
    assert model["p"].amount_known() == len(struct.get_type_interp("A"))




from ..._fodot.vocabulary import _bool, _int, _real, _parse_builtin_type  # type: ignore
from ..vocabulary import (
    BuiltinBool,
    BuiltinInt,
    BuiltinReal,
    IntType,
    RealType,
    StrType,
    Pfunc,
)
from enum import Enum, EnumType
from typing import TypeVar, Self, TypeAlias, Any
from abc import ABC


class ParseAble(ABC):
    @classmethod
    def parse(cls, _: str) -> Self: ... # type: ignore[empty-body]


_EnumMember = TypeVar("_EnumMember", bound=ParseAble)


class ParseAbleEnumType(EnumType, ParseAble):
    def __contains__(cls: type[_EnumMember], member: Any) -> bool: # type: ignore[misc]
        if isinstance(member, str):
            try:
                member = cls.parse(member)
            except ValueError:
                pass
        return super().__contains__(member) # type: ignore[misc]

    def __getitem__(cls: type[_EnumMember], name: str) -> _EnumMember: # type: ignore[misc]
        try:
            return super().__getitem__(name)  # type: ignore
        except KeyError:
            return cls.parse(name)


class BuiltinTypes(Enum, metaclass=ParseAbleEnumType):
    """
    An `enum.Enum` of builtin types available in all FO(·) vocabularies.

    Aquire the actual type values using the `value` attribute on the enum members.

    See also the python `Enum` documentation.

    # `__getitem__`

    Standard `__getitem__` behaviour on an `Enum` only recognizes the names of the enum members.
    This behaviour is extended to allow any string that can be parsed by `BuiltinTypes.parse`.

    ```python
    from sli_lib.fodot.vocabulary import BuiltinTypes
    assert BuiltinTypes.BOOL == BuiltinTypes["Bool"] == BuiltinTypes["𝔹"]
    ```

    # `__contains__`

    The behaviour of `__contains__` has been similarly extended to allow parseable strings.

    ```python
    from sli_lib.fodot.vocabulary import BuiltinTypes
    assert "Bool" in BuiltinTypes
    assert "𝔹" in BuiltinTypes
    ```
    """

    @classmethod
    def parse(cls, value: str) -> Self:
        """
        Parse the given value as a builtin type.
        """
        return cls(_parse_builtin_type(value))

    BOOL = _bool
    """
    Builtin `Bool` type in FO(·).

    This member is of type `BuiltinBool`.
    """
    INT = _int
    """
    Builtin `Int` type in FO(·).

    This member is of type `BuiltinInt`.
    """
    REAL = _real
    """
    Builtin `Real` type in FO(·).

    This member is of type `BuiltinReal`.
    """

    def __str__(self) -> str:
        return str(self.value)


BuiltinTypeAlias: TypeAlias = BuiltinTypes | BuiltinBool | BuiltinInt | BuiltinReal

Type: TypeAlias = BuiltinTypeAlias | IntType | RealType | StrType

CustomType: TypeAlias = IntType | RealType | StrType

ExtType: TypeAlias = Type | str

ExtCustomType: TypeAlias = CustomType | str

ExtPfunc: TypeAlias = Pfunc | str

__all__ = [
    "BuiltinTypes",
    "BuiltinTypeAlias",
    "Type",
    "CustomType",
    "ExtCustomType",
    "ExtType",
    "ExtPfunc",
]



from typing import Iterable, List, Any
from itertools import chain


def filter__all__(value: Iterable[str]) -> List[str]:
    return list(
        filter(
            lambda x: not x.startswith("_"),
            iter(value),
        )
    )


def merge__all__(*args: Iterable[str]) -> List[str]:
    return list(chain(*args))

def rename_module(all: Iterable[Any], local_values: dict, new_name: str) -> None:
    for type in all:
        try:
            local_values[type].__module__ = new_name
        except AttributeError:
            pass



from enum import Enum

class SatResult(Enum):
    SAT = "sat"
    UNSAT = "unsat"
    UNKNOWN = "unknown"

    def __bool__(self) -> bool:
        return self.value == SatResult.SAT.value

__all__ = [
    "SatResult",
]



from typing import TypeAlias, TypeVar, TypeGuard, Protocol, overload
from ...fodot.vocabulary import Real
from ..structure import (
    PfuncInterp, GlobModelPfuncInterp, ModelPfuncInterp,
    IntInterp, RealInterp, StrInterp, PfuncInterpIter,
    GlobModelPfuncInterpIter, ModelPfuncInterpIter
)

TypeElement: TypeAlias = bool | int | Real | str
ExtTypeElement: TypeAlias = TypeElement | float
ExtReal: TypeAlias = Real | int | float
TypeInterp: TypeAlias = IntInterp | RealInterp | StrInterp

T = TypeVar("T", bound=TypeElement, covariant=True)

class PfuncGeneric(Protocol[T]):
    """
    Private protocol for any generic class about pfunc interpretations.
    """
    def _py_codomain(self) -> type[T]: ...

@overload
def has_codomain(
    interp: PfuncInterp[TypeElement],
    codomain: type[T]
) -> TypeGuard[PfuncInterp[T]]:
    return issubclass(interp._py_codomain(), codomain)
@overload
def has_codomain(
    interp: GlobModelPfuncInterp[TypeElement],
    codomain: type[T]
) -> TypeGuard[GlobModelPfuncInterp[T]]:
    return issubclass(interp._py_codomain(), codomain)
@overload
def has_codomain(
    interp: ModelPfuncInterp[TypeElement],
    codomain: type[T]
) -> TypeGuard[ModelPfuncInterp[T]]:
    return issubclass(interp._py_codomain(), codomain)
@overload
def has_codomain(
    interp: PfuncInterpIter[TypeElement],
    codomain: type[T]
) -> TypeGuard[PfuncInterpIter[T]]:
    return issubclass(interp._py_codomain(), codomain)
@overload
def has_codomain(
    interp: GlobModelPfuncInterpIter[TypeElement],
    codomain: type[T]
) -> TypeGuard[GlobModelPfuncInterpIter[T]]:
    return issubclass(interp._py_codomain(), codomain)
@overload
def has_codomain(
    interp: ModelPfuncInterpIter[TypeElement],
    codomain: type[T]
) -> TypeGuard[ModelPfuncInterpIter[T]]:
    return issubclass(interp._py_codomain(), codomain)
def has_codomain(
    interp: PfuncGeneric[TypeElement],
    codomain: type[T]
) -> TypeGuard[PfuncGeneric[T]]:
    """
    Used for checking the type of the codomain of some pfunc related object.

    Returns true if `codomain` is a subclass of the pfunc's codomain, false otherwise.
    This function tells static type analyzers to narrow the type of the given class.
    Do note that this functions works with unions, but unions do not work for narrowing the type,
    this is a limitation of the python type system.

    # Example
     
    ```python
    from sli_lib.fodot.structure import (
        PfuncInterp, PfuncInterpIter,
        has_codomain,
    )
    from typing import reveal_type

    def pfunc_interp(p: PfuncInterp) -> None:
        if has_codomain(p, bool):
            # This will be `PfuncInterp[bool]` instead of `PfuncInterp[TypeElement]`
            reveal_type(p)

    def pfunc_interp_iter(p: PfuncInterpIter) -> None:
        if has_codomain(p, bool):
            # This will be `PfuncInterpIter[bool]` instead of `PfuncInterpIter[TypeElement]`
            reveal_type(p)
    ```

    """
    return issubclass(interp._py_codomain(), codomain)

__all__ = [
    "TypeElement",
    "ExtTypeElement",
    "ExtReal",
    "TypeInterp",
    "PfuncInterp",
    "has_codomain",
    "PfuncGeneric",
]



"""

# Introduction

sli-lib is a library for manipulating and using the FO(·) language using the rust sli-lib library.
sli-lib contains two major parts the `fodot` module for representing and manipulating FO(·) and the `solver` module for executing inference tasks on FO(·) theories.

# `fodot`

The `fodot` module contains three modules `vocabulary`, `theory` and `structure`.
Each containing datastructures and method for representing FO(·) constructs for vocabularies, theories and structures respectively.

# `solver`

In the `solver` module, solver implementations can be found and used with a `Theory` object from `fodot.theory`.
Each solver implements some inference tasks over this theory such as, satisfiability check, model expansion, back-bone propagation, ...

# Example

```python
from sli_lib.fodot import Vocabulary, Structure, Assertions, Theory
from sli_lib.fodot.structure import StrInterp
from sli_lib.solver import Z3Solver

# Create a vocabulary
vocab = Vocabulary("V")
vocab.add_type("Node")
vocab.add_type("Color")
vocab.add_pfunc("edge", ("Node", "Node"), "Bool")
vocab.add_pfunc("colorOf", ("Node",), "Color")
# Create some assertions
assertions = Assertions(vocab)
assertions.parse("!x, y in Node: edge(x, y) => colorOf(x) ~= colorOf(y).")
# Create a structure
struct = Structure(vocab)
struct.set_type_interp("Node", StrInterp(f"a{i}" for i in range(20)))
struct.set_type_interp("Color", StrInterp(["red", "green", "blue"]))
# Add some edges to struct
edge = struct["edge"]
for args in [
    ("a1", "a2"),
    ("a1", "a3"),
    ("a3", "a2"),
    ("a1", "a4"),
    ("a2", "a4"),
    ("a3", "a4"),
]:
    edge.set(args, True)
edge.set_all_unknown_to_value(False)
# Put our assertions and structure together in a Theory and ask for satisfiability from Z3Solver
theory = Theory(assertions, struct)
solver = Z3Solver(theory)
assert not solver.check() # this instance of graph colouring is unsatisfiable
```

# Free threading support

sli-lib does not rely on the Python global interpreter lock.
As such it fully supports free threaded Python.
Just like the builtin Python types, the sli-lib types, where needed, add locks to prevent memory unsafety and data races.
Do note that this requires building the api explicitly targeting free-threaded python.

"""

from .sli_lib import * # type: ignore[import-not-found]
from . import fodot as fodot
from . import solver as solver

# Keep this up to date manually
__all__ = [
    "fodot",
    "solver",
]



from ._pure import *
from ..vocabulary import Real, Pfunc, Vocabulary, ExtType, ExtPfunc
from typing import Iterable, Self, TypeVar, Generic

T = TypeVar("T", bound=TypeElement)

class PfuncInterpIter(Generic[T]):
    "" # workaround for pdoc priority of stub docs picking up docs for Generic[T]
    def __iter__(self) -> Self: ...
    def __next__(self) -> tuple[tuple[TypeElement, ...], T]: ...
    def _py_codomain(self) -> type[T]: ...

class PfuncInterp(Generic[T]):
    ""
    def get(self, *args: ExtTypeElement) -> T | None: ...
    def set(self, args: tuple[ExtTypeElement, ...], value: T | None) -> None: ...
    def set_if_unknown(self, args: tuple[ExtTypeElement, ...], value: T) -> bool: ...
    def set_all_unknown_to_value(self, value: T) -> None: ...
    def any_known(self) -> bool: ...
    def amount_known(self) -> int: ...
    def amount_unknown(self) -> int: ...
    def is_complete(self) -> bool: ...
    def __call__(self, *args: ExtTypeElement) -> T | None: ...
    def decl(self) -> Pfunc: ...
    def vocab(self) -> Vocabulary: ...
    def structure(self) -> "Structure": ...
    def name(self) -> str: ...
    def _py_codomain(self) -> type[T]: ...
    def __str__(self) -> str: ...
    def __iter__(self) -> PfuncInterpIter[T]: ...

class GlobModelPfuncInterpIter(Generic[T]):
    ""
    def __iter__(self) -> Self: ...
    def __next__(self) -> tuple[tuple[TypeElement, ...], T]: ...
    def _py_codomain(self) -> type[T]: ...

class GlobModelPfuncInterp(Generic[T]):
    ""
    def get(self, *args: ExtTypeElement) -> T | None: ...
    def any_known(self) -> bool: ...
    def amount_known(self) -> int: ...
    def amount_unknown(self) -> int: ...
    def is_complete(self) -> bool: ...
    def __call__(self, *args: ExtTypeElement) -> T | None: ...
    def decl(self) -> Pfunc: ...
    def vocab(self) -> Vocabulary: ...
    def glob_model(self) -> "GlobModel": ...
    def name(self) -> str: ...
    def _py_codomain(self) -> type[T]: ...
    def __str__(self) -> str: ...
    def __iter__(self) -> GlobModelPfuncInterpIter[T]: ...

class ModelPfuncInterpIter(Generic[T]):
    ""
    def __iter__(self) -> Self: ...
    def __next__(self) -> tuple[tuple[TypeElement, ...], T]: ...
    def _py_codomain(self) -> type[T]: ...

class ModelPfuncInterp(Generic[T]):
    ""
    def get(self, *args: ExtTypeElement) -> T: ...
    def __call__(self, *args: ExtTypeElement) -> T: ...
    def decl(self) -> Pfunc: ...
    def vocab(self) -> Vocabulary: ...
    def model(self) -> "Model": ...
    def name(self) -> str: ...
    def _py_codomain(self) -> type[T]: ...
    def __str__(self) -> str: ...
    def __iter__(self) -> ModelPfuncInterpIter[T]: ...

class IntInterpIter:
    def __iter__(self) -> Self: ...
    def __next__(self) -> int: ...

class IntInterp:
    def __init__(self, values: Iterable[int] | None=None): ...
    def __str__(self) -> str: ...
    def __repr__(self) -> str: ...
    def __len__(self) -> int: ...
    def __iter__(self) -> IntInterpIter: ...

class RealInterpIter:
    def __iter__(self) -> Self: ...
    def __next__(self) -> Real: ...

class RealInterp:
    def __init__(self, values: Iterable[ExtReal] | None=None): ...
    def __str__(self) -> str: ...
    def __repr__(self) -> str: ...
    def __len__(self) -> int: ...
    def __iter__(self) -> RealInterpIter: ...

class StrInterpIter:
    def __iter__(self) -> Self: ...
    def __next__(self) -> str: ...

class StrInterp:
    def __init__(self, values: Iterable[str] | None=None): ...
    def __str__(self) -> str: ...
    def __repr__(self) -> str: ...
    def __len__(self) -> int: ...
    def __iter__(self) -> StrInterpIter: ...

class Structure:
    def __init__(self, vocabulary: Vocabulary): ...
    def __getitem__(self, name: ExtPfunc) -> PfuncInterp[TypeElement]: ...
    def get_type_interp(self, type: ExtType) -> TypeInterp | None: ...
    def set_type_interp(self, type: ExtType, type_interp: TypeInterp) -> None: ...
    def completed_domain(self) -> bool: ...
    def is_complete(self) -> bool: ...
    def parse(self, decls: str) -> "Structure": ...
    def vocab(self) -> Vocabulary: ...
    def iter_complete(self) -> "CompleteStructureIter": ...
    def __str__(self) -> str: ...

class CompleteStructureIter:
    def enable_skip_infinite(self) -> None: ...
    def disable_skip_infinite(self) -> None: ...
    def skip_infinite(self, skip: bool) -> None: ...
    def __iter__(self) -> Self: ...
    def __next__(self) -> Structure: ...

class GlobModel:
    def __getitem__(self, pfunc: ExtPfunc) -> GlobModelPfuncInterp[TypeElement]: ...
    def vocab(self) -> Vocabulary: ...
    def __str__(self) -> str: ...
    def iter_models(self) -> "CompleteModelIter": ...
    def into_structure(self) -> Structure: ...

class CompleteModelIter:
    def enable_skip_infinite(self) -> None: ...
    def disable_skip_infinite(self) -> None: ...
    def skip_infinite(self, skip: bool) -> None: ...
    def __iter__(self) -> Self: ...
    def __next__(self) -> "Model": ...

class Model:
    def __getitem__(self, pfunc: ExtPfunc) -> ModelPfuncInterp[TypeElement]: ...
    def vocab(self) -> Vocabulary: ...
    def __str__(self) -> str: ...
    def into_structure(self) -> Structure: ...



from mypy import api
from sys import stderr, stdout

def test_typing():
    out, err, exit_code = api.run(["--disallow-untyped-defs", "-p", "sli_lib"])
    print(out, file=stdout)
    print(err, file=stderr)
    assert exit_code == 0



"""

# Subscripting (`__getitem__`)

All structure like classes (`Structure`, `GlobModel`, ...) can be subscripted (`structure["p"]`) with an `..vocabulary.ExtPfunc` to acquire the interpretation of that pfunc, or raise an exception if this pfunc doesn't exist in the vocabulary or the structure.
"""

from ..._fodot.structure import *
from ..._fodot import structure

# Don't enumerate each thing we want to import here do it in __init__.pyi
from ._pure import *
from . import _pure
from ..._all_utils import filter__all__, merge__all__, rename_module

__all__ = merge__all__(
    filter__all__(structure.__all__),
    _pure.__all__,
)

# patch __module__ of everything to this module
rename_module(__all__, locals(), __name__)
